EV_TO_NM = 1239.841  # energy -> wavelength (freq)
AU_TO_EV = 27.211324570273  # wavelength (freq) -> enegy
DEBYE_TO_AU = 0.393430  # dipole moment
ANGSTROM_TO_AU = 1.889716
AMU_TO_ELECTRONMASS = 1822.8884843  # mass
BOHR_TO_CM = 0.00000000529177249
SPEEDOFLIGHT_AU = 137.0359996  # c in au
KB_EV = 8.61735E-05  # Boltzmann constant eV/k
EV_TO_CM = 8065.5401069 # eV to cm-1 (freq)